/*
 * Copyright (C) 2019  CZ.NIC, z. s. p. o.
 *
 * This file is part of FRED.
 *
 * FRED is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * FRED is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with FRED.  If not, see <https://www.gnu.org/licenses/>.
 */

#include "include/libpg/util/strong_type_operators.hh"

#include <iostream>
#include <type_traits>
#include <utility>

namespace {

template <typename T>
constexpr T increment(T value)
{
    return ++value;
}

template <typename T>
constexpr T post_increment(T value)
{
    return value++;
}

template <typename T>
constexpr T decrement(T value)
{
    return --value;
}

template <typename T>
constexpr T post_decrement(T value)
{
    return value--;
}

template <typename, typename = void> struct IsIncrementable : std::false_type {};
template <typename T>
struct IsIncrementable<T, std::enable_if_t<sizeof(++std::declval<T&>()) != 0>> : std::true_type {};

template <typename, typename = void> struct IsPostIncrementable : std::false_type {};
template <typename T>
struct IsPostIncrementable<T, std::enable_if_t<sizeof(std::declval<T&>()++) != 0>> : std::true_type {};

template <typename, typename = void> struct IsDecrementable : std::false_type {};
template <typename T>
struct IsDecrementable<T, std::enable_if_t<sizeof(--std::declval<T&>()) != 0>> : std::true_type {};

template <typename, typename = void> struct IsPostDecrementable : std::false_type {};
template <typename T>
struct IsPostDecrementable<T, std::enable_if_t<sizeof(std::declval<T&>()--) != 0>> : std::true_type {};

template <typename, typename = void> struct IsStreamable : std::false_type {};
template <typename T>
struct IsStreamable<T, std::enable_if_t<std::is_same<decltype(std::declval<std::ostream&>() << std::declval<T>()), std::ostream&>::value>> : std::true_type {};

using NamedInt = LibPg::Util::StrongType<int, struct NamedIntTag_>;
static_assert(sizeof(NamedInt) == sizeof(int), "unexpected size of NamedInt type");
static_assert(NamedInt{0} == NamedInt{0}, "unexpected comparison result");
static_assert(!(NamedInt{0} == NamedInt{1}), "unexpected comparison result");
static_assert(NamedInt{0} != NamedInt{1}, "unexpected comparison result");
static_assert(!(NamedInt{0} != NamedInt{0}), "unexpected comparison result");
static_assert(NamedInt{0} < NamedInt{1}, "unexpected comparison result");
static_assert(!(NamedInt{0} < NamedInt{0}), "unexpected comparison result");
static_assert(!(NamedInt{0} < NamedInt{-1}), "unexpected comparison result");
static_assert(NamedInt{0} <= NamedInt{0}, "unexpected comparison result");
static_assert(NamedInt{0} <= NamedInt{1}, "unexpected comparison result");
static_assert(!(NamedInt{0} <= NamedInt{-1}), "unexpected comparison result");
static_assert(!(NamedInt{0} > NamedInt{1}), "unexpected comparison result");
static_assert(!(NamedInt{0} > NamedInt{0}), "unexpected comparison result");
static_assert(NamedInt{0} > NamedInt{-1}, "unexpected comparison result");
static_assert(!(NamedInt{0} >= NamedInt{1}), "unexpected comparison result");
static_assert(NamedInt{0} >= NamedInt{0}, "unexpected comparison result");
static_assert(NamedInt{0} >= NamedInt{-1}, "unexpected comparison result");
static_assert(!NamedInt{0}, "unexpected comparison result");
static_assert(!(!NamedInt{1}), "unexpected comparison result");
static_assert(NamedInt{0}.get() == 0, "unexpected value");
static_assert(increment(NamedInt{0}) == NamedInt{1}, "unexpected comparison result");
static_assert(post_increment(NamedInt{0}) == NamedInt{0}, "unexpected comparison result");
static_assert(decrement(NamedInt{1}) == NamedInt{0}, "unexpected comparison result");
static_assert(post_decrement(NamedInt{0}) == NamedInt{0}, "unexpected comparison result");
static_assert(-NamedInt{1} == NamedInt{-1}, "unexpected result");

static constexpr NamedInt zero{0};
static constexpr NamedInt one{1};
static_assert(!zero, "unexpected comparison result");
static_assert(!(!one), "unexpected comparison result");
static_assert(zero.get() == 0, "unexpected value");
static_assert(one.get() == 1, "unexpected value");

struct X {};
static constexpr X x{};
static_assert(sizeof(X) == 1, "unexpected size of X type");
static_assert(sizeof(x) == sizeof(X), "unexpected size of X instance");
using NamedX = LibPg::Util::StrongType<X, struct NamedXTag_>;
static constexpr NamedX named_x{x};
static_assert(sizeof(named_x) == sizeof(X), "unexpected size of NamedX instance");
static_assert(IsIncrementable<int>::value, "operator++(int&) must exist");
static_assert(IsPostIncrementable<int>::value, "operator++(int&, int) must exist");
static_assert(IsDecrementable<int>::value, "operator--(int&) must exist");
static_assert(IsPostDecrementable<int>::value, "operator--(int&, int) must exist");
static_assert(IsIncrementable<NamedInt>::value, "operator++(NamedInt&) must exist");
static_assert(IsPostIncrementable<NamedInt>::value, "operator++(NamedInt&, int) must exist");
static_assert(IsDecrementable<NamedInt>::value, "operator--(NamedInt&) must exist");
static_assert(IsPostDecrementable<NamedInt>::value, "operator--(NamedInt&, int) must exist");
static_assert(!IsIncrementable<X>::value, "operator++(X&) can not exist");
static_assert(!IsPostIncrementable<X>::value, "operator++(X&, int) can not exist");
static_assert(!IsDecrementable<X>::value, "operator--(X&) can not exist");
static_assert(!IsPostDecrementable<X>::value, "operator--(X&, int) can not exist");
static_assert(!IsIncrementable<NamedX>::value, "operator++(NamedX&) can not exist");
static_assert(!IsPostIncrementable<NamedX>::value, "operator++(NamedX&, int) can not exist");
static_assert(!IsDecrementable<NamedX>::value, "operator--(NamedX&) can not exist");
static_assert(!IsPostDecrementable<NamedX>::value, "operator--(NamedX&, int) can not exist");

static_assert(IsStreamable<int>::value, "operator<<(std::ostream&, const int&) must exist");
static_assert(IsStreamable<NamedInt>::value, "operator<<(std::ostream&, const NamedInt&) must exist");
static_assert(!IsStreamable<X>::value, "operator<<(std::ostream&, const X&) can not exist");
static_assert(!IsStreamable<NamedX>::value, "operator<<(std::ostream&, const NamedX&) can not exist");

struct Y
{
    Y operator++(int);
    Y& operator--();
};
using NamedY = LibPg::Util::StrongType<Y, struct NamedYTag_>;
static_assert(!IsIncrementable<Y>::value, "operator++(Y&) can not exist");
static_assert(IsPostIncrementable<Y>::value, "operator++(Y&, int) must exist");
static_assert(IsDecrementable<Y>::value, "operator--(Y&) must exist");
static_assert(!IsPostDecrementable<Y>::value, "operator--(Y&, int) can not exist");
static_assert(!IsIncrementable<NamedY>::value, "operator++(NamedY&) can not exist");
static_assert(IsPostIncrementable<NamedY>::value, "operator++(NamedY&, int) must exist");
static_assert(IsDecrementable<NamedY>::value, "operator--(NamedY&) must exist");
static_assert(!IsPostDecrementable<NamedY>::value, "operator--(NamedY&, int) can not exist");

}//namespace {anonymous}
